# Exception Clasess
class InstallPandasException(Exception):
    pass


class APIColumnNameError(Exception):
    pass


class InvalidFrequencyError(Exception):
    pass


class MissingRequiredArgumentError(Exception):
    pass
